import mqr.inference.lib.rate as rate
import mqr.inference.lib.util as util

from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest
from mqr.inference.power import TestPower

import mqr.interop.inference as interop

import numpy as np
import scipy
import statsmodels

import warnings

def power_1sample(ra, H0_rate, nobs, alpha, meas=1.0, alternative='two-sided', method='norm-approx'):
    '''
    Calculate power of a test of rate of events.

    Null-hypothesis
        `ra` / `H0_rate` == 1

    Parameters
    ----------
    ra : float
        Alternative hypothesis rate, forming effect size.
    H0_rate : float
        Null-hypothesis rate.
    nobs : int
        Number of observations.
    alpha : float
        Required significance.
    meas : float, optional
        Extent of one period in observation.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'norm-approx'}, optional
        Test method. Only 'norm-approx', the normal approximation to
        the binomial distribution, is available.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    if method == 'norm-approx':
        dist = scipy.stats.norm()
        var_a = ra / (nobs * meas)
        var_0 = H0_rate / (nobs * meas)
        den = np.sqrt(var_a)

        if alternative == 'less':
            z = dist.ppf(1-alpha)
            num = z * np.sqrt(var_0) + ra - H0_rate
            power = 1 - dist.cdf(num/den)
        elif alternative == 'greater':
            z = dist.ppf(1-alpha)
            num = z * np.sqrt(var_0) + H0_rate - ra
            power = 1 - dist.cdf(num/den)
        elif alternative == 'two-sided':
            z = dist.ppf(1-alpha/2)
            num_1 = z * np.sqrt(var_0) + H0_rate - ra
            num_2 = H0_rate - z * np.sqrt(var_0) - ra
            power = 1 - (dist.cdf(num_1/den) - dist.cdf(num_2/den))
        else:
            raise ValueError(util.alternative_error_msg(alternative))
    else:
        raise ValueError(util.method_error_msg(method, ['norm-approx']))

    return TestPower(
        name='rate of events',
        alpha=alpha,
        beta=1-power,
        effect=f'{ra:g} / {H0_rate:g} = {ra/H0_rate:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs)

def power_2sample(r1, r2, nobs, alpha, H0_value=None, meas1=1.0, meas2=1.0,
                  alternative='two-sided', method='score', compare='diff'):
    '''
    Calculate power of a test of difference or ratio of two rates of events.

    Null-hypothesis by compare
        | 'diff'
        |   `r1` - `r2` == `H0_value`
        | 'ratio'
        |   `r1` / `r2` == `H0_value`

    Parameters
    ----------
    r1 : float
        First rate.
    r2 : float
        Second rate.
    nobs : int
        Number of observations.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    H0_value : float, optional
        Null-hypothesis rate. Default 0 for 'diff', 1 for 'ratio'.
    alternative : {'two-sided', 'less', 'greater'}
        Sense of alternative hypothesis.
    method : str, optional
        Test method. See statsmodels documentation.
    compare : {'diff', 'ratio'}, optional
        | 'diff'
        |   Compares the difference between rates.
            Calls :func:`sm..power_poisson_diff_2indep <statsmodels.stats.rates.power_poisson_diff_2indep>`.
        | 'ratio'
        |   Compares the ratio of two rates.
            Calls :func:`sm..power_poisson_ratio_2indep <statsmodels.stats.rates.power_poisson_ratio_2indep>`.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    alt = interop.alternative(alternative, lib='statsmodels')
    if compare == 'diff':
        desc = 'difference between'
        sample_stat_sym = '-'
        sample_stat_value = r1 - r2
        if H0_value is None:
            H0_value = 0.0
        power = statsmodels.stats.rates.power_poisson_diff_2indep(
            rate1=r1,
            rate2=r2,
            nobs1=nobs,
            nobs_ratio=1,
            value=H0_value,
            alpha=alpha,
            alternative=alt,
            method_var=method,
            return_results=False)
    elif compare == 'ratio':
        desc = 'ratio of'
        sample_stat_sym = '/'
        sample_stat_value = r1 / r2
        if H0_value is None:
            H0_value = 1.0
        power = statsmodels.stats.rates.power_poisson_ratio_2indep(
            rate1=r1,
            rate2=r2,
            nobs1=nobs,
            nobs_ratio=1,
            value=H0_value,
            alpha=alpha,
            alternative=alt,
            method_var=method,
            return_results=False)
    else:
        raise ValueError(util.compare_error_msg('test'))
    return TestPower(
        name=f'{desc} rates of events',
        alpha=alpha,
        beta=1-power,
        effect=f'{r1/meas1:g} {sample_stat_sym} {r2/meas2:g} = {sample_stat_value:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs)

def size_1sample(ra, H0_rate, alpha, beta, meas=1.0, alternative='two-sided', method='norm-approx'):
    '''
    Calculate sample size for test of rate of events.

    Null-hypothesis
        `ra` / `H0_rate` == 1

    Parameters
    ----------
    ra : float
        Alternative hypothesis rate, forming effect size.
    H0_rate : float
        Null-hypothesis rate.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'norm-approx', 'chi2'}, optional
        | 'norm-approx'
        |   Approximation to the normal distribution. Numerically solves
            :func:`power_1sample` equal to 1 - `beta` by searching `nobs`.
        | 'chi2'
        |   Numerically finds the sample size that produces the requested power
            using a Chi-squared distribution.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    if method == 'chi2':
        n = 1
        r = ra / H0_rate if ra > H0_rate else H0_rate / ra
        if alternative == 'less' or alternative == 'greater':
            NP1 = 1 - beta
            DP1 = alpha
        elif alternative == 'two-sided':
            NP1 = 1 - beta
            DP1 = alpha / 2
        else:
            raise ValueError(util.alternative_error_msg(alternative))
        def ratio(n):
            num = scipy.stats.chi2.ppf(NP1, df=n)
            den = scipy.stats.chi2.ppf(DP1, df=n)
            return num / den - r
        df_opt = scipy.optimize.fsolve(ratio, 1)[0]
        num = scipy.stats.chi2.ppf(NP1, df=df_opt)
        nobs_opt = num / 2.0 / np.maximum(ra, H0_rate)
    elif method == 'norm-approx':
        def beta_fn(n):
            return power_1sample(
                ra=ra,
                H0_rate=H0_rate,
                nobs=n,
                alpha=alpha,
                meas=meas,
                alternative=alternative).beta - beta
        nobs_opt = scipy.optimize.fsolve(beta_fn, 1)[0]
    else:
        raise ValueError(util.method_error_msg(method, ['chi2', 'norm-approx']))

    return TestPower(
        name='rate of events',
        alpha=alpha,
        beta=beta,
        effect=f'{ra:g} / {H0_rate:g} = {ra/H0_rate:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs_opt,)

def size_2sample(r1, r2, alpha, beta, H0_value=None, alternative='two-sided', method='score', compare='diff'):
    '''
    Calculate sample size for difference of two rates of events.

    Null-hypothesis by compare
        | 'diff'
        |   `r1` - `r2` == `H0_value`
        | 'ratio'
        |   `r1` / `r2` == `H0_value`

    Parameters
    ----------
    r1 : float
        First rate.
    r2 : float
        Second rate.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    H0_value : float, optional
        Null-hypothesis rate. Default 0 for 'diff', 1 for 'ratio'.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : str, optional
        | 'z'
        |   Approximation based a normal approximation. For this method,
            `compare` must be 'diff' and `H0_value` must be 0.
        | (other)
        |   All other methods are passed as parameters to :func:`power_2sample`
            while that function is solved numerically equal to the requested
            power by varying `nobs`.
    compare : {'diff', 'ratio'}, optional
        | 'diff'
        |   Compare the difference between rates.
        | 'ratio'
        |   Compare the ratio of rates.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    if H0_value is None:
        if compare == 'diff':
            H0_value = 0.0
        elif compare == 'ratio':
            H0_value = 1.0

    if method == 'z':
        if compare != 'diff':
            raise ValueError(f'comparison "{compare}" not available with `z` method')
        if not np.isclose(H0_value, 0):
            raise ValueError(f'H0_value "{H0_value}" must be 0 with `z` method')
        if alternative != 'two-sided':
            crit = alpha
        else:
            crit = alpha / 2
        Zb = scipy.stats.norm().ppf(1-beta)
        Za = -scipy.stats.norm().ppf(crit)
        num = Za * np.sqrt(r2) + Zb * np.sqrt(r1)
        nobs_opt = 2 * num**2 / (r1 - r2)**2
    else:
        def beta_fn(nobs):
            power = power_2sample(
                r1,
                r2,
                nobs,
                alpha,
                H0_value=H0_value,
                alternative=alternative,
                method=method,
                compare=compare)
            return power.beta - beta
        nobs_opt = scipy.optimize.fsolve(beta_fn, 2)[0]

    if compare == 'diff':
        desc = 'difference between'
        sample_stat_sym = '-'
        sample_stat_value = r1 - r2
    elif compare == 'ratio':
        desc = 'ratio of'
        sample_stat_sym = '/'
        sample_stat_value = r1 / r2
    else:
        raise ValueError(util.compare_error_msg(compare))

    return TestPower(
        name=f'{desc} rates of events',
        alpha=alpha,
        beta=beta,
        effect=f'{r1:g} {sample_stat_sym} {r2:g} = {sample_stat_value:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs_opt)

def confint_1sample(count, n, meas=1.0, conf=0.95, bounded='both', method='wald-cc'):
    '''
    Confidence interval for rate `count / n / meas`.

    Parameters
    ----------
    count : int
        Number of events.
    n : int
        Number of periods over which events were counted.
    meas : float, optional
        Extent of one period of observation.
    conf : float, optional
        Confidence level that determines the width of the interval.
    method : {'chi2', 'exact', 'wald', 'wald-cc'}, optional
        | 'chi2'
        |   Chi2 interval, see [2]_.
        | 'exact'
        |   Exact method, recommended for small `count`.
            Implements method 9 in [3]_.
        | 'wald-cc'
        |   Wald method with continuity correction, recommended for small `count`.
            Implements method 5 in [1]_.
        | (other)
        |   Everything else is passed to
            :func:`sm..confint_poisson <statsmodels.stats.rates.confint_poisson>`,
            which supports only two-sided intervals.

    Notes
    -----
    For a discussion on the benefits and disadvantages of various intervals,
    including these, see [1]_ and [3]_.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Patil, V. V., & Kulkarni, H. V. (2012).
            Comparison of confidence intervals for the Poisson mean: some new aspects.
            REVSTAT-Statistical Journal, 10(2), 211-22.
    .. [2]  Garwood, F. (1936).
            Fiducial limits for the Poisson distribution.
            Biometrika, 28(3/4), 437-442.
    .. [3]  Barker, L. (2002).
            A comparison of nine confidence intervals for a Poisson parameter when
            the expected number of events is ≤ 5.
            The American Statistician, 56(2), 85-89.
    '''
    value = count / n / meas
    alpha = 1 - conf
    if method == 'chi2':
        if count <= 4:
            msg = (
                'This method is recommended when `count > 4`. '
                'Consider using method "exact" or "wald-cc". See [1].')
            warnings.warn(msg)
        (lower, upper) = rate.confint_1sample_chi2(count, n, meas, conf, bounded)
    elif method == 'exact':
        (lower, upper) = rate.confint_1sample_exact(count, n, meas, conf, bounded)
    elif method == 'wald-cc':
        (lower, upper) = rate.confint_1sample_wald_cc(count, n, meas, conf, bounded)
    else:
        (lower, upper) = statsmodels.stats.rates.confint_poisson(
            count=count,
            exposure=n*meas,
            method=method,
            alpha=alpha)
    return ConfidenceInterval(
        name='rate of events',
        method=method,
        value=value,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def confint_2sample(count1, n1, count2, n2, meas1=1.0, meas2=1.0,
                    conf=0.95, bounded='both', method='wald'):
    '''
    Confidence interval for difference of rates `count1 / n1 / meas1 - count2 / n2 / meas2`.

    Parameters
    ----------
    count1 : int
        Number of events in first observation.
    n1 : int
        Number of periods over which first events were counted.
    count2 : int
        Number of events in second observation.
    n2 : int
        Number of periods over which second events were counted.
    meas1 : float, optional
        Extent of one period in first observation.
    meas2 : float, optional
        Extent of one period in second observation.
    conf : float, optional
        Confidence level that determines the width of the interval.
    method : str, optional
        | 'wald'
        |   Wald's normal approximation, equation (9) in [1]_.
        | 'wald-moment'
        |   Based on Wald, equation (8) in [1]_. Use for small/zero counts.
        | (other)
        |   Everything else is passed to
            :func:`sm..confint_poisson_2indep <statsmodels.stats.rates.confint_poisson_2indep>`,
            which supports only two-sided intervals.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Krishnamoorthy, K., & Lee, M. (2013).
            New approximate confidence intervals for the difference between
            two Poisson means and comparison.
            Journal of Statistical Computation and Simulation, 83(12), 2232-2243.
    '''
    if method == 'wald':
        (lower, upper) = rate.confint_2sample_wald(
            count1, n1, count2, n2,
            meas1, meas2,
            conf, bounded)
    elif method == 'wald-moment':
        (lower, upper) = rate.confint_2sample_wald_moment(
            count1, n1, count2, n2,
            meas1, meas2,
            conf, bounded)
    else:
        if bounded != 'both':
            raise ValueError(
                f'Method "{method}" is passed to statsmodels which does not implement '
                'one-sided bounds. ' +
                util.method_error_msg(method, ['wald', 'wald-moment']))
        alpha = 1 - conf
        (lower, upper) = statsmodels.stats.rates.confint_poisson_2indep(
            count1=count1,
            exposure1=n1*meas1,
            count2=count2,
            exposure2=n2*meas2,
            method=method,
            compare='diff',
            alpha=alpha)

    value = count1 / n1 / meas1 - count2 / n2 / meas2
    return ConfidenceInterval(
        name='difference between rates of events',
        method=method,
        value=value,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def test_1sample(count, n, meas=1.0, H0_rate=1.0, alternative='two-sided', method='exact-c'):
    '''
    Hypothesis test for the rate of events.

    Null-hypothesis
        `count` / `n` / `meas` == `H0_rate`

    Calls :func:`sm..test_poisson <statsmodels.stats.rates.test_poisson>`.

    Parameters
    ----------
    count : int
        Number of events.
    n : int
        Number of periods over which events were counted.
    meas : float, optional
        Extent of one period of observation.
    H0_rate : float, optional
        Null-hypothesis rate.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : str, optional
        Test method. See statsmodels docs.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    '''
    alt = interop.alternative(alternative, lib='statsmodels')
    res = statsmodels.stats.rates.test_poisson(
        count=count,
        nobs=n*meas,
        value=H0_rate,
        method=method,
        alternative=alt,)

    return HypothesisTest(
        description='rate of events',
        alternative=alternative,
        method=method,
        sample_stat=f'rate',
        sample_stat_target=H0_rate,
        sample_stat_value=count/n/meas,
        stat=res.statistic,
        pvalue=res.pvalue,)

def test_2sample(count1, n1, count2, n2, meas1=1.0, meas2=1.0,
                 H0_value=None, alternative='two-sided', method='score', compare='diff'):
    '''
    Hypothesis test for equality of rates.

    Null-hypothesis by compare
        | 'diff'
        |   `count1` / `n1` / `meas1` - `count2` / `n2` / `meas2` == `H0_value`
        | 'ratio'
        |   `count1` / `n1` / `meas1` / (`count2` / `n2` / `meas2`) == `H0_value`

    Calls :func:`sm..test_poisson_2indep <statsmodels.stats.rates.test_poisson_2indep>`.

    Parameters
    ----------
    count1 : int
        Number of events in first observation.
    n1 : int
        Number of periods over which first events were counted.
    count2 : int
        Number of events in second observation.
    n2 : int
        Number of periods over which second events were counted.
    meas1 : float, optional
        Extent of one period in first observation.
    meas2 : float, optional
        Extent of one period in second observation.
    H0_value : float, optional
        Null-hypothesis value. Default 1 when compare is 'ratio', 0 when compare
        is 'diff'.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : str, optional
        See statsmodels docs.
    compare : {'diff', 'ratio'}, optional
        | 'diff'
        |   Compare the difference between rates.
        | 'ratio'
        |   Compare the rate of ratios.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    '''
    alt = interop.alternative(alternative, lib='statsmodels')
    res = statsmodels.stats.rates.test_poisson_2indep(
        count1=count1,
        exposure1=n1*meas1,
        count2=count2,
        exposure2=n2*meas2,
        value=H0_value,
        method=method,
        compare=compare,
        alternative=alt)

    if compare == 'diff':
        desc = 'difference between'
        sample_stat_sym = '-'
        sample_stat_value = count1 / n1 / meas1 - count2 / n2 / meas2
        if H0_value is None:
            H0_value = 0.0
    elif compare == 'ratio':
        desc = 'ratio of'
        sample_stat_sym = '/'
        sample_stat_value = count1 * n2 * meas2 / (count2 * n1 * meas1)
        if H0_value is None:
            H0_value = 1.0
    else:
        raise ValueError(util.compare_error_msg(compare))

    return HypothesisTest(
        description=f'{desc} rates of events',
        alternative=alternative,
        method=method,
        sample_stat=f'rate1 {sample_stat_sym} rate2',
        sample_stat_target=H0_value,
        sample_stat_value=sample_stat_value,
        stat=res.statistic,
        pvalue=res.pvalue,)



"""
======================================
SPC alarm rules (:mod:`mqr.spc.rules`)
======================================

.. currentmodule:: mqr.spc.rules

User guide
    :doc:`/user_guide/statistical-process-control`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

.. rubric:: Functions

.. autosummary::
    :toctree: generated/

    limits
    aofb_nsigma
    n_1side
    n_trending
    combine
"""

from dataclasses import dataclass, field
import functools
import numpy as np
import pandas as pd

def combine(combn_fn, *rules):
    """
    Create a new rule from others and a logical combination.

    **A word on types**. This function can be confusing because it transforms
    functions into another function. The list of arguments `rules` and the
    result of this function all have the same signature. Take a look at
    :func:`limits` for an example of that signature. On the other hand,
    `combn_fn` has a different signature: it takes two series of bools and
    returns another series of bools whose elements correspond to the first two.
    That is, it takes results from two rules and makes a new result.
    See the example below.

    The new rule is the same as evaluating each rule in `rules`, then combining
    the outputs with `combn_fn(left_result, right_result)` from left to right.

    Parameters
    ----------
    combn_fn : Callable[[[Series[bool], pandas.Series[bool]], pandas.Series[bool]]
        Function that combines the result series from two rules into a single
        Series.
    *rules : Callable[[ControlStatistic, ControlParams], Series[bool]]
        One or more rules whose results will be combined.

    Examples
    --------

    **Creating a combination**

    This example creates a combination function that alarms when both of its
    inputs are alarmed, but not when only one is an alarm.

    >>> def both(series1, series2):
    >>>     result = pd.Series(index=series1.index, dtype=bool)
    >>>     for i in range(len(result)):
    >>>         result[i] = series1[i] and series2[i]
    >>>     return result

    That code demonstrates the interface with a simple example, but in practise,
    use :func:`numpy.logical_and` for the functionality of "both" above, since
    it is much more efficient, and already written:

    >>> mqr.spc.rules.combine(np.logical_and, rule1, rule2)

    **Using a combination**

    Create a rule that alarms when either (1) a process goes outside its control
    limits using :func:`mqr.spc.rules.limits()`, or (2) the process has 3 of 4
    observations outside 2 sigmas using :func:`mqr.spc.rules.aofb_nsigma`. To
    make the numbers simple, the process here has standard deviation of 1 and
    samples of size 1 are taken -- a single observation of 1 is then a sample
    with mean at 1 standard error.

    1. The data first violates the UCL of 3, with an average observation 3.5.
    2. The data then violates the 3/4 outside 2sigma limit, with the four sample
       averages 2.5, 1, 2.5, 2.5 (three of which are bigger than 2).
       The last of these four observations is labelled as violating the limit.

    .. plot::

        fig, ax = plt.subplots(figsize=(7, 3))

        # Make the new rule
        combine_rule = mqr.spc.rules.combine(
            np.logical_or,
            mqr.spc.rules.limits(),
            mqr.spc.rules.aofb_nsigma(a=3, b=4, n=2))

        # Create violating data for demonstration
        mean = 5
        sigma = 1
        zscore = np.array([[0, 3.5, 0, 0, 0, 2.5, 1, 2.5, 2.5, 0]]).T
        #                      ^^^(>UCL)     ^^^^^^^^^^^^^^^^(3/4>2)
        data = pd.DataFrame(mean + zscore * sigma)

        # Create parameters and calculate statistic for the example chart
        params = mqr.spc.XBarParams(centre=mean, sigma=sigma)
        stat = params.statistic(data)

        # Show the chart and alarms overlay for the combined rule
        mqr.plot.spc.chart(stat, params, ax=ax)
        mqr.plot.spc.alarms(stat, params, combine_rule, ax=ax)

    """
    def _combine(control_statistic, control_params):
        alarms = [rule(control_statistic, control_params) for rule in rules]
        return functools.reduce(combn_fn, alarms)
    return _combine

def limits():
    """
    Rule monitoring when a process violates control limits.

    This function is an alarm rule that triggers an alarm any time the statistic
    is outside the upper or lower control limits.

    This is a rule and can be used directly in :func:`mqr.plot.spc.alarms` or
    :func:`combine`.

    Returns
    -------
    Callable[(ControlStatistic, ControlParams), pandas.Series[bool]]
        A function taking a control statistic and the params used to create them,
        and returning a series with `True` marking alarms.

    Examples
    --------
    This function can be passed directly to the plotting routines. This example
    overlays alarms on an X-bar chart.

    .. plot::

        data = pd.DataFrame({
            'Xbar': np.array([9, 10, 9, 13, 12, 10])
        #                               ^^ (>UCL)
        })

        params = mqr.spc.XBarParams(centre=10, sigma=1)
        stat = params.statistic(data)
        rule = mqr.spc.rules.limits()

        fig, ax = plt.subplots(figsize=(7, 3))
        mqr.plot.spc.chart(stat, params, ax=ax)
        mqr.plot.spc.alarms(stat, params, rule, ax=ax)

    """
    def _rule(control_statistic, control_params):
        stat = control_statistic.stat
        nobs = control_statistic.nobs
        lcl = control_params.lcl(nobs)
        ucl = control_params.ucl(nobs)
        return np.logical_or(stat >= ucl, stat <= lcl)
    return _rule

def aofb_nsigma(a, b, n):
    """
    Rule monitoring the proportion of samples outside a multiple of sigma.

    This function returns an alarm rule that alarms when `a` out of `b` statistics
    in a row are beyond `n` multiples of the statistic standard error. Only the
    last sample of the `b` is marked as an alarm.

    This function generates a rule. The result of this function is another
    function that can be passed to :func:`mqr.plot.spc.alarms` or
    :func:`combine`.

    Parameters
    ----------
    a : int
        Intensity -- the number of statistics in `b` periods outside `n`
        standard deviations required to trigger an alarm.
    b : int
        Period -- the number of samples to check for `a` statistics outside
        `n` standard deviations required to trigger an alarm.
    n : float
        Threshold in multiples of the standard error.

    Returns
    -------
    Callable[(ControlStatistic, ControlParams), pandas.Series[bool]]
        A function taking a control statistic and the params used to create them,
        and returning a series with `True` marking alarms.

    Examples
    --------
    The result of this function can be passed to the plotting routines. This
    example overlays a "4/5 > 2sigma" rule on an X-bar chart.

    .. plot::

        data = pd.DataFrame({
            'Xbar': np.array([11, 13, 10, 12, 12, 12, 9, 11, 12])
        #                         ^^^^^^^^^^^^^^^^^^ (4/5>=2sigma)
        })

        params = mqr.spc.XBarParams(centre=10, sigma=1)
        stat = params.statistic(data)
        rule = mqr.spc.rules.aofb_nsigma(a=4, b=5, n=2)

        fig, ax = plt.subplots(figsize=(7, 3))
        mqr.plot.spc.chart(stat, params, ax=ax)
        mqr.plot.spc.alarms(stat, params, rule, ax=ax)

    """
    if a > b:
        raise ValueError(f'Cannot detect more than b of b signals (was passed "{a} of {b}").')
    def _rule(control_statistic, control_params):
        stat = control_statistic.stat
        nobs = control_statistic.nobs
        target = control_params.target()
        se = control_params.se(nobs)

        alarms = pd.Series(False, index=stat.index)
        for seq in (stat >= target + n * se).rolling(b):
            if np.sum(seq) >= a:
                alarms[seq.index[-1]] = True
        for seq in (stat <= target - n * se).rolling(b):
            if np.sum(seq) >= a:
                alarms[seq.index[-1]] = True
        return alarms
    return _rule

def n_1side(n):
    """
    Rule monitoring the count on one side the centreline.

    This function returns an alarm rule that alarms when `n` statistics in a row
    fall on one side of the centreline.

    This function generates a rule. The result of this function is another
    function that can be passed to :func:`mqr.plot.spc.alarms` or
    :func:`combine`.

    Parameters
    ----------
    n : int
        Number of consecutive statistics falling on one side of the centreline
        required to trigger an alarm.

    Returns
    -------
    Callable[(ControlStatistic, ControlParams), pandas.Series[bool]]
        A function taking a control statistic and the params used to create them,
        and returning a series with `True` marking alarms.

    Examples
    --------

    .. plot::

        data = pd.DataFrame({
            'Xbar': np.array([9, 10, 11, 13, 12, 8])
        #                            ^^^^^^^^^^ (3 > target)
        })

        params = mqr.spc.XBarParams(centre=10, sigma=1)
        stat = params.statistic(data)
        rule = mqr.spc.rules.n_1side(3)

        fig, ax = plt.subplots(figsize=(7, 3))
        mqr.plot.spc.chart(stat, params, ax=ax)
        mqr.plot.spc.alarms(stat, params, rule, ax=ax)

    """
    def _rule(control_statistic, control_params):
        stat = control_statistic.stat
        target = control_params.target()

        alarms = pd.Series(False, index=stat.index)
        for seq in np.sign(stat - target).rolling(n):
            values = set(seq)
            if (len(seq) == n) and (len(values) == 1) and (values != {0}):
                alarms[seq.index[-1]] = True

        return alarms
    return _rule

def n_trending(n):
    """
    Rule monitoring the count of trending points.

    This function returns an alarm rule that alarms when `n` statistics in a row
    trend (up or down).

    This function generates a rule. The result of this function is another
    function that can be passed to :func:`mqr.plot.spc.alarms` or
    :func:`combine`.

    Parameters
    ----------
    n : int
        Number of sample in a trend required to cause an alarm.

    Returns
    -------
    Callable[(ControlStatistic, ControlParams), pandas.Series[bool]]
        A function taking a control statistic and the params used to create them,
        and returning a series with `True` marking alarms.

    Examples
    --------

    .. plot::

        data = pd.DataFrame({
            'Xbar': np.array([11, 9, 12, 10, 9, 8, 12])
        #                            ^^^^^^^^^^^^ (4 trending)
        })

        params = mqr.spc.XBarParams(centre=10, sigma=1)
        stat = params.statistic(data)
        rule = mqr.spc.rules.n_trending(4)

        fig, ax = plt.subplots(figsize=(7, 3))
        mqr.plot.spc.chart(stat, params, ax=ax)
        mqr.plot.spc.alarms(stat, params, rule, ax=ax)

    """
    def _rule(control_statistic, control_params):
        stat = control_statistic.stat
        target = control_params.target()

        alarms = pd.Series(False, index=stat.index)
        for seq in np.sign(stat.diff()).rolling(n-1):
            values = set(seq)
            if (len(seq) == n-1) and (len(values) == 1) and (values != {0}):
                alarms[seq.index[-1]] = True

        return alarms
    return _rule



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest
import scipy

import mqr

def test_test_1_sample():
    np.random.seed(1234)
    x = np.array([1, 2, 3, 3, 4, 4, 4, 5, 5, 7, 9])

    test = 'ad-norm'
    res = mqr.inference.dist.test_1sample(x, test)
    assert res.description == 'non-normality'
    assert res.alternative == 'two-sided'
    assert res.method == 'anderson-darling'
    assert res.sample_stat == 'distribution'
    assert res.sample_stat_target == 'normal'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    test = 'ks-norm'
    res = mqr.inference.dist.test_1sample(x, test)
    assert res.description == 'non-normality'
    assert res.alternative == 'two-sided'
    assert res.method == 'kolmogorov-smirnov'
    assert res.sample_stat == 'distribution'
    assert res.sample_stat_target == 'normal'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    test = 'lf-norm'
    res = mqr.inference.dist.test_1sample(x, test)
    assert res.description == 'non-normality'
    assert res.alternative == 'two-sided'
    assert res.method == 'lilliefors'
    assert res.sample_stat == 'distribution'
    assert res.sample_stat_target == 'normal'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    test = 'sw-norm'
    res = mqr.inference.dist.test_1sample(x, test)
    assert res.description == 'non-normality'
    assert res.alternative == 'two-sided'
    assert res.method == 'shapiro-wilk'
    assert res.sample_stat == 'distribution'
    assert res.sample_stat_target == 'normal'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    test = 'dp-norm'
    x_large = scipy.stats.norm().rvs(30) # Uses kurtosistest, which warns for n<20
    res = mqr.inference.dist.test_1sample(x_large, test)
    assert res.description == 'non-normality'
    assert res.alternative == 'two-sided'
    assert res.method == 'dagostino-pearson'
    assert res.sample_stat == 'distribution'
    assert res.sample_stat_target == 'normal'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    test = 'jb-norm'
    with pytest.warns(UserWarning):
        res = mqr.inference.dist.test_1sample(x, test)
        assert res.pvalue > 0.05 # Only a small-sample warning
        assert res.description == 'non-normality'
        assert res.alternative == 'two-sided'
        assert res.method == 'jarque-bera'
        assert res.sample_stat == 'distribution'
        assert res.sample_stat_target == 'normal'
        assert res.sample_stat_value == None
        assert isinstance(res.stat, numbers.Number)
        assert isinstance(res.pvalue, numbers.Number)

    x_non_norm = scipy.stats.gamma(2).rvs(500)
    with pytest.warns(UserWarning):
        res = mqr.inference.dist.test_1sample(x_non_norm, test)
        assert res.pvalue < 0.05 # Only a small-pvalue warning, since n=500
        assert res.description == 'non-normality'
        assert res.alternative == 'two-sided'
        assert res.method == 'jarque-bera'
        assert res.sample_stat == 'distribution'
        assert res.sample_stat_target == 'normal'
        assert res.sample_stat_value == None
        assert isinstance(res.stat, numbers.Number)
        assert isinstance(res.pvalue, numbers.Number)

    x_large = scipy.stats.norm().rvs(500)
    res = mqr.inference.dist.test_1sample(x_large, test)
    assert res.description == 'non-normality'
    assert res.alternative == 'two-sided'
    assert res.method == 'jarque-bera'
    assert res.sample_stat == 'distribution'
    assert res.sample_stat_target == 'normal'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    test = 'cvm-norm'
    res = mqr.inference.dist.test_1sample(x, test)
    assert res.description == 'non-normality'
    assert res.alternative == 'two-sided'
    assert res.method == 'cramer-vonmises'
    assert res.sample_stat == 'distribution'
    assert res.sample_stat_target == 'normal'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    with pytest.raises(Exception) as e_info:
        mqr.inference.dist.test_1sample(x, 'not-a-test')
    assert e_info.value.args[0] == 'Method "not-a-test" is not available. Use ad-norm, ks-norm, lf-norm, sw-norm, dp-norm, jb-norm or cvm-norm.'



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest

import mqr

def test_size_1sample():
    effect = 1.0
    alpha = 0.01
    beta = 0.01
    alternative = 'two-sided'

    res = mqr.inference.mean.size_1sample(effect, alpha, beta, alternative)
    assert res.name == 'mean'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.method == 't'
    assert res.alternative == alternative
    assert isinstance(res.sample_size, numbers.Number)

def test_size_2sample():
    effect = 1.0
    alpha = 0.01
    beta = 0.01
    alternative = 'two-sided'

    res = mqr.inference.mean.size_2sample(effect, alpha, beta, alternative)
    assert res.name == 'difference between means (independent)'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.method == 't'
    assert res.alternative == alternative
    assert isinstance(res.sample_size, numbers.Number)

def test_size_paired():
    effect = 1.0
    alpha = 0.01
    beta = 0.01
    alternative = 'two-sided'

    res = mqr.inference.mean.size_paired(effect, alpha, beta, alternative)
    assert res.name == 'difference between means (paired)'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.method == 't'
    assert res.alternative == alternative
    assert isinstance(res.sample_size, numbers.Number)

def test_confint_1sample():
    x = np.array([0.9, 1.1])
    conf = 0.90
    bounded = 'both'
    alternative = 'both'

    method = 't'
    res = mqr.inference.mean.confint_1sample(x, conf, alternative, method)
    assert res.name == 'mean'
    assert res.method == method
    assert res.conf == conf
    assert res.bounded == bounded
    assert res.value == np.mean(x)
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)

    method = 'z'
    res = mqr.inference.mean.confint_1sample(x, conf, alternative, method)
    assert res.name == 'mean'
    assert res.method == method
    assert res.conf == conf
    assert res.bounded == bounded
    assert res.value == np.mean(x)
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)

def test_confint_2sample():
    x = np.array([0.9, 1.1])
    y = np.array([1.9, 2.1])
    conf = 0.90
    pooled = True
    bounded = 'both'

    method = 't'
    res = mqr.inference.mean.confint_2sample(x, y, conf, pooled, bounded, method)
    assert res.name == 'difference between means (independent)'
    assert res.method == method
    assert conf == conf
    assert res.bounded == bounded
    assert res.value == np.mean(x) - np.mean(y)
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)

    method = 'z'
    res = mqr.inference.mean.confint_2sample(x, y, conf, pooled, bounded, method)
    assert res.name == 'difference between means (independent)'
    assert res.method == method
    assert conf == conf
    assert res.bounded == bounded
    assert res.value == np.mean(x) - np.mean(y)
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)

def test_confint_paired():
    x = np.array([0.9, 1.1])
    y = np.array([1.9, 2.1])
    conf = 0.90
    bounded = 'both'

    method = 't'
    res = mqr.inference.mean.confint_paired(x, y, conf, bounded, method)
    assert res.name == 'difference between means (paired)'
    assert res.method == method
    assert conf == conf
    assert res.bounded == bounded
    assert res.value == np.mean(x) - np.mean(y)
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)

    method = 'z'
    res = mqr.inference.mean.confint_paired(x, y, conf, bounded, method)
    assert res.name == 'difference between means (paired)'
    assert res.method == method
    assert conf == conf
    assert res.bounded == bounded
    assert res.value == np.mean(x) - np.mean(y)
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)

def test_test_1sample():
    x = np.array([0.9, 1.1])
    H0_mean = 1.1
    alternative = 'two-sided'

    res = mqr.inference.mean.test_1sample(x, H0_mean, alternative, 'z')
    assert res.description == 'mean'
    assert res.alternative == alternative
    assert res.method == 'z'
    assert res.sample_stat == 'mean(x)'
    assert res.sample_stat_target == H0_mean
    assert res.sample_stat_value == np.mean(x)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    res = mqr.inference.mean.test_1sample(x, H0_mean, 'two-sided', 't')
    assert res.description == 'mean'
    assert res.alternative == alternative
    assert res.method == 't'
    assert res.sample_stat == 'mean(x)'
    assert res.sample_stat_target == H0_mean
    assert res.sample_stat_value == np.mean(x)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

def test_test_2sample():
    x = np.array([0.9, 1.1])
    y = np.array([1.9, 2.1])
    H0_diff = 1.1
    pooled = True
    alternative = 'two-sided'

    res = mqr.inference.mean.test_2sample(x, y, H0_diff, pooled, alternative, 't')
    assert res.description == 'difference between means (independent)'
    assert res.alternative == alternative
    assert res.method == 't'
    assert res.sample_stat == 'mean(x) - mean(y)'
    assert res.sample_stat_target == H0_diff
    assert res.sample_stat_value == np.mean(x) - np.mean(y)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    res = mqr.inference.mean.test_2sample(x, y, H0_diff, pooled, alternative, 'z')
    assert res.description == 'difference between means (independent)'
    assert res.alternative == alternative
    assert res.method == 'z'
    assert res.sample_stat == 'mean(x) - mean(y)'
    assert res.sample_stat_target == H0_diff
    assert res.sample_stat_value == np.mean(x) - np.mean(y)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

def test_test_paired():
    x = np.array([0.9, 1.1])
    y = np.array([1.8, 2.2])
    alternative = 'two-sided'
    method = 't'

    res = mqr.inference.mean.test_paired(x, y, alternative, method)
    assert res.description == 'difference between means (paired)'
    assert res.alternative == alternative
    assert res.method == 't'
    assert res.sample_stat == 'mean(x) - mean(y)'
    assert res.sample_stat_target == 0.0
    assert res.sample_stat_value == np.mean(x) - np.mean(y)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)



import numbers
import numpy as np
import pytest

import mqr

def test_size_1sample():
    alpha = 0.10
    beta = 0.10

    effect = 1.5
    alternative = 'two-sided'
    res = mqr.inference.variance.size_1sample(effect, alpha, beta, alternative)
    assert res.name == 'variance'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert np.ceil(res.sample_size) == 105

    alternative = 'greater'
    res = mqr.inference.variance.size_1sample(effect, alpha, beta, alternative)
    assert res.name == 'variance'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert np.ceil(res.sample_size) == 82

    effect = 0.8
    alternative = 'less'
    res = mqr.inference.variance.size_1sample(effect, alpha, beta, alternative)
    assert res.name == 'variance'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == effect
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert np.ceil(res.sample_size) == 266

def test_size_2sample():
    alpha = 0.10
    beta = 0.10

    var_ratio = 1.5
    alternative = 'two-sided'
    res = mqr.inference.variance.size_2sample(var_ratio, alpha, beta, alternative)
    assert res.name == 'ratio of variances'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == var_ratio
    assert res.alternative == alternative
    assert res.method == 'f'
    assert np.ceil(res.sample_size) == 211

    alternative = 'greater'
    res = mqr.inference.variance.size_2sample(var_ratio, alpha, beta, alternative)
    assert res.name == 'ratio of variances'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == var_ratio
    assert res.alternative == alternative
    assert res.method == 'f'
    assert np.ceil(res.sample_size) == 162

    var_ratio = 0.8
    alternative = 'less'
    res = mqr.inference.variance.size_2sample(var_ratio, alpha, beta, alternative)
    assert res.name == 'ratio of variances'
    assert res.alpha == alpha
    assert res.beta == beta
    assert res.effect == var_ratio
    assert res.alternative == alternative
    assert res.method == 'f'
    assert np.ceil(res.sample_size) == 530

def test_confint_1sample():
    x = np.array([ 98.89,  98.75,  97.61, 100.38,  99.56,
                  103.12,  97.73, 102.39, 96.74, 100.06])
    s2 = np.var(x, ddof=1)
    conf = 0.95

    bounded = 'both'
    res = mqr.inference.variance.confint_1sample(x, conf, bounded=bounded)
    assert res.name == 'variance'
    assert res.method == 'chi2'
    assert res.value == s2
    assert res.lower == pytest.approx(1.9871, 1e-4)
    assert res.upper == pytest.approx(13.9980, 1e-4)
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'below'
    res = mqr.inference.variance.confint_1sample(x, conf, bounded=bounded)
    assert res.name == 'variance'
    assert res.method == 'chi2'
    assert res.value == s2
    assert res.lower == pytest.approx(2.2342, 1e-4)
    assert res.upper == np.inf
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'above'
    res = mqr.inference.variance.confint_1sample(x, conf, bounded=bounded)
    assert res.name == 'variance'
    assert res.method == 'chi2'
    assert res.value == s2
    assert res.lower == 0.0
    assert res.upper == pytest.approx(11.3680, 1e-4)
    assert res.conf == conf
    assert res.bounded == bounded

def test_confint_2sample():
    x = np.array([0, 1, 2])
    y = np.array([0, 2, 4])
    ratio = np.var(x, ddof=1) / np.var(y, ddof=1)
    conf = 0.90

    bounded = 'both'
    res = mqr.inference.variance.confint_2sample(x, y, conf, bounded=bounded)
    assert res.name == 'ratio of variances'
    assert res.method == 'f'
    assert res.value == ratio
    assert res.lower == pytest.approx(0.0132, abs=1e-4)
    assert res.upper == pytest.approx(4.75)
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'below'
    res = mqr.inference.variance.confint_2sample(x, y, conf, bounded=bounded)
    assert res.name == 'ratio of variances'
    assert res.method == 'f'
    assert res.value == ratio
    assert res.lower == pytest.approx(0.0278, abs=1e-4)
    assert res.upper == np.inf
    assert res.conf == conf
    assert res.bounded == bounded

    bounded = 'above'
    res = mqr.inference.variance.confint_2sample(x, y, conf, bounded=bounded)
    assert res.name == 'ratio of variances'
    assert res.method == 'f'
    assert res.value == ratio
    assert res.lower == -np.inf
    assert res.upper == pytest.approx(2.25)
    assert res.conf == conf
    assert res.bounded == bounded

def test_test_1sample():
    x = np.array([0, 1, 2])
    H0_var = 1

    alternative = 'two-sided'
    res = mqr.inference.variance.test_1sample(x, H0_var, alternative)
    assert res.description == 'variance'
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert res.sample_stat == 'var(x)'
    assert res.sample_stat_target == H0_var
    assert res.sample_stat_value == np.var(x, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.7358, abs=1e-4)

    alternative = 'greater'
    res = mqr.inference.variance.test_1sample(x, H0_var, alternative)
    assert res.description == 'variance'
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert res.sample_stat == 'var(x)'
    assert res.sample_stat_target == H0_var
    assert res.sample_stat_value == np.var(x, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.3679, abs=1e-4)

    alternative = 'less'
    res = mqr.inference.variance.test_1sample(x, H0_var, alternative)
    assert res.description == 'variance'
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert res.sample_stat == 'var(x)'
    assert res.sample_stat_target == H0_var
    assert res.sample_stat_value == np.var(x, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.6321, abs=1e-4)

def test_test_2sample():
    x = np.array([0, 1, 2])
    y = np.array([1, 2, 4])

    method = 'f'
    alternative = 'two-sided'
    res = mqr.inference.variance.test_2sample(x, y, alternative, method)
    assert res.description == 'ratio of variances'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'var(x) / var(y)'
    assert res.sample_stat_target == 1.0
    assert res.sample_stat_value == np.var(x, ddof=1) / np.var(y, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.6)

    alternative = 'less'
    res = mqr.inference.variance.test_2sample(x, y, alternative, method)
    assert res.description == 'ratio of variances'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'var(x) / var(y)'
    assert res.sample_stat_target == 1.0
    assert res.sample_stat_value == np.var(x, ddof=1) / np.var(y, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.3)

    alternative = 'greater'
    res = mqr.inference.variance.test_2sample(x, y, alternative, method)
    assert res.description == 'ratio of variances'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'var(x) / var(y)'
    assert res.sample_stat_target == 1.0
    assert res.sample_stat_value == np.var(x, ddof=1) / np.var(y, ddof=1)
    assert isinstance(res.stat, numbers.Number)
    assert res.pvalue == pytest.approx(0.7)

    method = 'levene'
    alternative = 'two-sided'
    res = mqr.inference.variance.test_2sample(x, y, alternative, method)
    assert res.description == 'equality of variances'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'var(x) - var(y)'
    assert res.sample_stat_target == 0.0
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    method = 'bartlett'
    res = mqr.inference.variance.test_2sample(x, y, alternative, method)
    assert res.description == 'equality of variances'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'var(x) - var(y)'
    assert res.sample_stat_target == 0.0
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)



"""
======================================
Plotting tools (:mod:`mqr.plot.tools`)
======================================

.. currentmodule:: mqr.plot.tools

.. rubric:: Functions
.. autosummary::
    :toctree: generated/

    sharex
    sharey
"""

def sharey(fig, axs, remove_space=True):
    for ax in axs:
        if ax != axs[0]:
            ax.sharey(axs[0])
        ax.label_outer(remove_inner_ticks=remove_space)
        if remove_space:
            fig.get_layout_engine().set(w_pad=0, wspace=0)

def sharex(fig, axs, remove_space=True):
    for ax in axs:
        if ax != axs[0]:
            ax.sharex(axs[0])
        ax.label_outer(remove_inner_ticks=remove_space)
        if remove_space:
            fig.get_layout_engine().set(h_pad=0, hspace=0)



"""
============================================
Statistical Process Control (:mod:`mqr.spc`)
============================================

.. :currentmodule: mqr.spc

User guide
    :doc:`/user_guide/statistical-process-control`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide


This module provides classes for statistical process control. It is designed for
use with :mod:`mqr.plot.spc`, which provides plotting routines for the types
defined in this module. The module can also be used independently of control
charts, for example in scripts that detect alarm conditions and take some action.

.. rubric:: Classes

These are superclasses that define the behaviour of chart parameters and statistics.

.. autosummary::
    :toctree: generated/

    ControlParams
    ShewhartParams
    ControlStatistic

These are subclasses of :class:`ControlParams`. Users can define new control
parameters by subclassing :class:`ControlParams` or :class:`ShewhartParams`.

.. autosummary::
    :toctree: generated/

    XBarParams
    RParams
    SParams
    EwmaParams
    MewmaParams

.. rubric:: Submodules

These submodules define alarm rules and other utilities, including unbiasing constants.

.. autosummary::

    rules
    util

"""

from mqr.spc.lib.control import ControlParams, ControlStatistic, ShewhartParams
from mqr.spc.lib.control import XBarParams, RParams, SParams, EwmaParams, MewmaParams
from mqr.spc.util import solve_arl, solve_h4

import mqr.spc.rules
import mqr.spc.util



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest
import scipy

import mqr

def test_test_1sample():
    x = np.array([1, 6, 2, 3, 4, 9, 7, 3, 3 ,4])
    alternative = 'two-sided'
    
    method = 'runs'
    res = mqr.inference.nonparametric.dist.test_1sample(x, method)
    assert res.description == 'randomness'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'dist(x)'
    assert res.sample_stat_target == 'iid'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

def test_test_2sample():
    x = np.array([1, 2, 7, 3, 4, 5, 5, 4, 7, 9], dtype=float)
    y = np.array([9, 3, 6, 4, 1, 7, 7, 4, 5, 3], dtype=float)

    alternative = 'less'
    method = 'ks'
    res = mqr.inference.nonparametric.dist.test_2sample(x, y, alternative, method)
    assert res.description == 'sampling distribution'
    assert res.alternative == alternative
    assert res.method == 'kolmogorov-smirnov'
    assert res.sample_stat == 'dist(x)'
    assert res.sample_stat_target == 'dist(y)'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    alternative = 'two-sided'
    method = 'runs'
    res = mqr.inference.nonparametric.dist.test_2sample(x, y, alternative, method)
    assert res.description == 'sampling distribution'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'dist(x)'
    assert res.sample_stat_target == 'dist(y)'
    assert res.sample_stat_value == None
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)
